/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef DISTRIBUTED_MATRIX_HPP
#define DISTRIBUTED_MATRIX_HPP

#include <cstddef>
#include <random>
#include <functional>
#include "DenseMPI/StrumpackDensePackage.hpp"
#include "CompressedSparseMatrix.hpp"
#include "MPI_wrapper.hpp"
#include "Random_wrapper.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixMPI;
  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixHSSMPI;
  template<typename scalar_t,typename real_t,typename integer_t> class FrontalMatrixDenseMPI;

  inline int indxl2g(int INDXLOC, int NB, int IPROC, int ISRCPROC, int NPROCS) {
    return NPROCS*NB*((INDXLOC-1)/NB) + (INDXLOC-1) % NB +
      ((NPROCS+IPROC-ISRCPROC) % NPROCS)*NB + 1;
  }
  inline int indxg2l(int INDXGLOB, int NB, int IPROC, int ISRCPROC, int NPROCS) {
    return NB*((INDXGLOB-1)/(NB*NPROCS)) + (INDXGLOB-1) % NB + 1;
  }
  inline int indxg2p(int INDXGLOB, int NB, int IPROC, int ISRCPROC, int NPROCS) {
    return ( ISRCPROC + (INDXGLOB - 1) / NB ) % NPROCS;
  }

  // make the blocksize a template parameter? better for performance if power of 2?
  template<typename scalar_t,typename real_t>
  class DistributedMatrix {
  public:

#ifdef STRUMPACK_PBLAS_BLOCKSIZE
    static const int default_MB = STRUMPACK_PBLAS_BLOCKSIZE;
    static const int default_NB = STRUMPACK_PBLAS_BLOCKSIZE;
#else
    static const int default_MB = 32;
    static const int default_NB = 32;
#endif

    DistributedMatrix();
    DistributedMatrix(const DistributedMatrix<scalar_t,real_t>& m);
    DistributedMatrix(int blacs_ctxt, int M, int N);
    DistributedMatrix(int blacs_ctxt, int M, int N, scalar_t* A, int all_ctxt);
    DistributedMatrix(int blacs_ctxt, int M, int N, int MB, int NB);
    DistributedMatrix(int desc[9]);
    virtual ~DistributedMatrix();
    virtual DistributedMatrix<scalar_t,real_t>& operator=(const DistributedMatrix<scalar_t,real_t>& m);
    virtual DistributedMatrix<scalar_t,real_t>& operator=(DistributedMatrix<scalar_t,real_t>&& m);
    inline int rows() { return _desc[2]; }
    inline int cols() { return _desc[3]; }
    inline int local_rows() { return loc_rows; }
    inline int local_cols() { return loc_cols; }
    inline int ld() { return loc_rows; } // leading dimension of local storage
    inline int MB() { return _desc[4]; }
    inline int NB() { return _desc[5]; }
    inline int local_row_blocks() { return std::ceil(float(local_rows()) / MB()); }
    inline int local_col_blocks() { return std::ceil(float(local_cols()) / NB()); }
    inline int mpi_comm() { return Cblacs2sys_handle(ctxt()); }
    inline int proc_row() { return p_row; }
    inline int proc_col() { return p_col; }
    inline int proc_rows() { return np_rows; }
    inline int proc_cols() { return np_cols; }
    inline int procs() { return (proc_rows() == -1) ? 0 : proc_rows()*proc_cols(); }
    inline bool ismaster() { return proc_row() == 0 && proc_col() == 0; }
    inline int rowl2g(int row) { assert(p_row != -1); return indxl2g(row+1, MB(), proc_row(), 0, proc_rows()) - 1; }
    inline int coll2g(int col) { assert(p_col != -1); return indxl2g(col+1, NB(), proc_col(), 0, proc_cols()) - 1; }
    inline int rowg2l(int row) { assert(p_row != -1); return indxg2l(row+1, MB(), proc_row(), 0, proc_rows()) - 1; }
    inline int colg2l(int col) { assert(p_col != -1); return indxg2l(col+1, NB(), proc_col(), 0, proc_cols()) - 1; }
    inline int rowg2p(int row) { assert(p_row != -1); return indxg2p(row+1, MB(), proc_row(), 0, proc_rows()); }
    inline int colg2p(int col) { assert(p_col != -1); return indxg2p(col+1, NB(), proc_col(), 0, proc_cols()); }
    inline int rank(int r, int c) { return rowg2p(r) + colg2p(c) * proc_rows(); }

    inline bool fixed() { return MB()==default_MB && NB()==default_NB; }
    inline int rowl2g_fixed(int row) { assert(p_row != -1); assert(fixed()); return indxl2g(row+1, default_MB, proc_row(), 0, proc_rows()) - 1; }
    inline int coll2g_fixed(int col) { assert(p_col != -1); assert(fixed()); return indxl2g(col+1, default_NB, proc_col(), 0, proc_cols()) - 1; }
    inline int rowg2l_fixed(int row) { assert(p_row != -1); assert(fixed()); return indxg2l(row+1, default_MB, proc_row(), 0, proc_rows()) - 1; }
    inline int colg2l_fixed(int col) { assert(p_col != -1); assert(fixed()); return indxg2l(col+1, default_NB, proc_col(), 0, proc_cols()) - 1; }
    inline int rowg2p_fixed(int row) { assert(p_row != -1); assert(fixed()); return indxg2p(row+1, default_MB, proc_row(), 0, proc_rows()); }
    inline int colg2p_fixed(int col) { assert(p_col != -1); assert(fixed()); return indxg2p(col+1, default_NB, proc_col(), 0, proc_cols()); }
    inline int rank_fixed(int r, int c) { assert(fixed()); return rowg2p_fixed(r) + colg2p_fixed(c) * proc_rows(); }

    inline scalar_t* data() { return _data; }
    inline int* desc() { return _desc; }
    inline bool active() { return p_row != -1; } // is this rank part of the blacs grid?
    inline int ctxt() { return _desc[1]; }
    inline scalar_t& operator()(int r, int c) { return _data[r+ld()*c]; }
    inline scalar_t& global(int r, int c) { return _data[rowg2l(r)+ld()*colg2l(c)]; }
    inline void set_global(int r, int c, scalar_t v) {
      if (rowg2p(r) == proc_row() && colg2p(c) == proc_col())
	global(r, c) = v;
    }
    inline void clear();
    inline void zero();
    inline void fill(scalar_t v);
    inline void random();
    inline void eye();
    inline bool isnan() {
      for (size_t i=0; i<loc_rows*loc_cols; i++) if (std::isnan(std::abs(_data[i]))) return true;
      return false;
    }
    inline bool checknan(std::string mat_name, int precision=15) {
      if (isnan()) { print(mat_name, precision); abort(); }
    }
    void print(std::string mat_name, int precision=15);
    inline void set_sequential(scalar_t* a, int all_ctxt);
    inline void get_sequential(scalar_t* a, int all_ctxt);
    inline void LU(int* ipiv);
    inline void getrs(char trans, DistributedMatrix<scalar_t,real_t>& B, int* ipiv);
    inline void laswp(int* ipiv);
    inline void lapiv(int* ipiv, DistributedMatrix<scalar_t,real_t>& A);


    using DM_t = DistributedMatrix<scalar_t,real_t>;
    template<typename integer_t>
    void extend_add_copy_to_buffers(DistributedMatrix<scalar_t,real_t>& F11,
				    DistributedMatrix<scalar_t,real_t>& F12,
				    DistributedMatrix<scalar_t,real_t>& F21,
				    DistributedMatrix<scalar_t,real_t>& F22,
				    std::vector<std::vector<scalar_t>>& sbuf,
				    FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>* pa,
				    std::vector<integer_t>& pa_ind);
    template<typename integer_t>
    void extend_add_column_copy_to_buffers(DistributedMatrix<scalar_t,real_t>& Bsep,
					   DistributedMatrix<scalar_t,real_t>& Bupd,
					   std::vector<std::vector<scalar_t>>& sbuf,
					   FrontalMatrixMPI<scalar_t,real_t,integer_t>* pa,
					   std::vector<integer_t>& pa_ind);
    template<typename integer_t>
    void skinny_extend_add_copy_to_buffers(DistributedMatrix<scalar_t,real_t>& S,
					   std::vector<std::vector<scalar_t>>& sbuf,
					   FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>* pa,
					   std::vector<integer_t>& pa_ind);
    template<typename _scalar_t,typename _real_t,typename integer_t>
    friend void extract_b_copy_to_buffers(DistributedMatrix<_scalar_t,_real_t>& Bsep,
					  DistributedMatrix<_scalar_t,_real_t>& Bupd,
					  std::vector<std::vector<_scalar_t>>& sbuf,
					  std::function<int(integer_t)> ch_rank,
					  std::vector<integer_t>& pa_ind, int ch_proc_rows);
    template<typename integer_t>
    void extract_copy_to_buffers(std::vector<integer_t>& rows, std::vector<integer_t>& cols,
				 std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
				 DistributedMatrix<scalar_t,real_t>& mat,
				 std::vector<std::vector<scalar_t>>& sbuf);
    template<typename integer_t>
    void extend_copy_to_buffers(std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
				DistributedMatrix<scalar_t,real_t>& mat,
				std::vector<std::vector<scalar_t>>& sbuf);
    template<typename _scalar_t,typename _real_t,typename integer_t>
    friend void extend_add_copy_from_buffers(DistributedMatrix<_scalar_t,_real_t>& F11,
					     DistributedMatrix<_scalar_t,_real_t>& F12,
					     DistributedMatrix<_scalar_t,_real_t>& F21,
					     DistributedMatrix<_scalar_t,_real_t>& F22,
					     _scalar_t** pbuf,
					     integer_t sep_begin, integer_t* pa_upd,
					     integer_t* ch_upd, integer_t ch_dim_upd,
					     std::function<int(integer_t,integer_t)> F22_child_rank);
    template<typename integer_t>
    void skinny_extend_add_copy_from_buffers(scalar_t** pbuf, integer_t sep_begin, integer_t dim_sep,
					     integer_t* pa_upd, integer_t* ch_upd, integer_t ch_dim_upd,
					     std::function<int(integer_t,integer_t)> chSrank);
    template<typename _scalar_t,typename _real_t,typename integer_t>
    friend void extend_add_column_copy_from_buffers(DistributedMatrix<_scalar_t,_real_t>& Bsep,
						    DistributedMatrix<_scalar_t,_real_t>& Bupd,
						    std::vector<std::vector<_scalar_t>>& buf,
						    integer_t sep_begin, integer_t* pa_upd,
						    integer_t* ch_upd, integer_t ch_dim_upd,
						    std::function<int(integer_t,integer_t)> b_child_rank);
    template<typename integer_t>
    void extract_b_copy_from_buffers(std::vector<std::vector<scalar_t>>& buf,
				     std::vector<integer_t>& pa_ind,
				     std::function<int(integer_t)> src_rank);
    template<typename integer_t>
    void copy_from_buffer(scalar_t** pbuf, integer_t* ch_upd, integer_t ch_dim_upd,
			  std::function<int(integer_t,integer_t)> F22_child_rank,
			  std::function<integer_t(integer_t)> front_row_to_global,
			  std::function<integer_t(integer_t)> front_col_to_global);
    template<typename integer_t>
    void copy_column_from_buffer(std::vector<scalar_t*>& pbuf, integer_t* ch_upd, integer_t ch_dim_upd,
				 std::function<int(integer_t,integer_t)> b_child_rank,
				 std::function<integer_t(integer_t)> front_to_global);
    template<typename integer_t>
    void extract_copy_from_buffers(std::vector<integer_t>& rows, std::vector<integer_t>& cols,
				   std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
				   DistributedMatrix<scalar_t,real_t>& mat,
				   scalar_t** pbuf);
    template<typename integer_t>
    void extend_copy_from_buffers(std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
				  DistributedMatrix<scalar_t,real_t>& mat, scalar_t** pbuf);

    template<typename integer_t>
    void fill_random(std::function<integer_t(integer_t)> global_row, integer_t dimA);

    template<typename _scalar_t,typename _real_t>
    friend void gemm(char transA, char transB, _scalar_t alpha, DistributedMatrix<_scalar_t,_real_t>& A,
		     DistributedMatrix<_scalar_t,_real_t>& B, _scalar_t beta,
		     DistributedMatrix<_scalar_t,_real_t>& C);
    template<typename _scalar_t,typename _real_t>
    friend void trsm(char side, char uplo, char transA, char diag, _scalar_t alpha,
		     DistributedMatrix<_scalar_t,_real_t>& A, DistributedMatrix<_scalar_t,_real_t>& B);
    template<typename _scalar_t,typename _real_t>
    friend void trsv(char uplo, char transA, char diag, DistributedMatrix<_scalar_t,_real_t>& A,
		     DistributedMatrix<_scalar_t,_real_t>& B);
    template<typename _scalar_t,typename _real_t>
    friend void gemv(char transA, _scalar_t alpha, DistributedMatrix<_scalar_t,_real_t>& A,
		     DistributedMatrix<_scalar_t,_real_t>& B, _scalar_t beta,
		     DistributedMatrix<_scalar_t,_real_t>& C);
    template<typename _scalar_t,typename _real_t>
    friend void geadd(char trans, _scalar_t alpha, DistributedMatrix<_scalar_t,_real_t>& A,
		      _scalar_t beta, DistributedMatrix<_scalar_t,_real_t>& C);

    template<typename integer_t>
    void extract_F11(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t sep_begin, integer_t dim_sep);
    template<typename integer_t>
    void extract_F12(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t upd_row_begin,
		     integer_t upd_col_begin, integer_t dim_upd, integer_t* upd);
    template<typename integer_t>
    void extract_F21(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t upd_row_begin,
		     integer_t upd_col_begin, integer_t dim_upd, integer_t* upd);

#if defined(WRITE_ROOT)
    void MPI_binary_write();
#endif

  protected:
    scalar_t* _data = NULL;
    int _desc[9];

    int loc_rows;
    int loc_cols;
    int np_rows;
    int np_cols;
    int p_row;
    int p_col;
  };

  /**
   * Wrapper class does exactly the same as a regular DistributedMatrix,
   * but it is initialized with existing memory, so it does not
   * allocate, own or delete the memory
   */
  template<typename scalar_t,typename real_t>
  class DistributedMatrixWrapper : public DistributedMatrix<scalar_t,real_t> {
  public:
    DistributedMatrixWrapper() : DistributedMatrix<scalar_t,real_t>() {}
    DistributedMatrixWrapper(int blacs_ctxt, int M, int N, scalar_t* A);
    DistributedMatrixWrapper(int blacs_ctxt, int M, int N, int MB, int NB, scalar_t* A);
    virtual ~DistributedMatrixWrapper();
    DistributedMatrixWrapper(const DistributedMatrix<scalar_t,real_t>& m);
    virtual DistributedMatrixWrapper& operator=(DistributedMatrix<scalar_t,real_t>& m);
    virtual DistributedMatrixWrapper& operator=(const DistributedMatrixWrapper<scalar_t,real_t>& m);
  };

  /**
   * Use the data pointed to by A, assume that data is already in 2D
   * block-cyclic layout.
   */
  template<typename scalar_t,typename real_t>
  DistributedMatrixWrapper<scalar_t,real_t>::DistributedMatrixWrapper
  (int blacs_ctxt, int M, int N, scalar_t* A) :
    DistributedMatrixWrapper(blacs_ctxt, M, N, this->default_MB, this->default_NB, A) {
  }

  /**
   * Use the data pointed to by A, assume that data is already in 2D
   * block-cyclic layout.
   */
  template<typename scalar_t,typename real_t>
  DistributedMatrixWrapper<scalar_t,real_t>::DistributedMatrixWrapper
  (int blacs_ctxt, int M, int N, int MB, int NB, scalar_t* A) {
    MB = std::max(1, MB);
    NB = std::max(1, NB);
    Cblacs_gridinfo(blacs_ctxt, &this->np_rows, &this->np_cols, &this->p_row, &this->p_col);
    if (this->p_row == -1 || this->p_col == -1) {
      this->loc_rows = this->loc_cols = 0;
      descset(this->_desc, M, N, MB, NB, 0, 0, blacs_ctxt, std::max(this->loc_rows,1));
      this->_data = NULL;
    } else {
      this->loc_rows = numroc(M, MB, this->p_row, 0, this->np_rows);
      this->loc_cols = numroc(N, NB, this->p_col, 0, this->np_cols);
      this->_data = A; // set _data pointer to existing memory A
      if (descinit(this->_desc, M, N, MB, NB, 0, 0, blacs_ctxt, std::max(this->loc_rows,1))) {
	std::cerr << "ERROR: Could not create DistributedMatrix descriptor!" << std::endl;
	exit(1);
      }
    }
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrixWrapper<scalar_t,real_t>::DistributedMatrixWrapper(const DistributedMatrix<scalar_t,real_t>& m) {
    *this = m;
  }

  template<typename scalar_t,typename real_t> DistributedMatrixWrapper<scalar_t,real_t>&
  DistributedMatrixWrapper<scalar_t,real_t>::operator=(const DistributedMatrixWrapper<scalar_t,real_t>& m) {
    this->np_rows = m.np_rows;   this->np_cols = m.np_cols;
    this->p_row = m.p_row;       this->p_col = m.p_col;
    this->loc_rows = m.loc_rows; this->loc_cols = m.loc_cols;
    for (int i=0; i<9; i++) this->_desc[i] = m._desc[i];
    delete[] this->_data;
    this->_data = m._data;
    return *this;
  }

  template<typename scalar_t,typename real_t> DistributedMatrixWrapper<scalar_t,real_t>&
  DistributedMatrixWrapper<scalar_t,real_t>::operator=(DistributedMatrix<scalar_t,real_t>& m) {
    this->np_rows = m.proc_rows();   this->np_cols = m.proc_cols();
    this->p_row = m.proc_row();      this->p_col = m.proc_col();
    this->loc_rows = m.local_rows(); this->loc_cols = m.local_cols();
    for (int i=0; i<9; i++) this->_desc[i] = m.desc()[i];
    delete[] this->_data;
    this->_data = m.data();
    return *this;
  }

  /**
   * destructor: do not call delete, the Wrapper class does not own the
   * data set NULL, so the parent destructor also does not delete the
   * data
   */
  template<typename scalar_t,typename real_t>
  DistributedMatrixWrapper<scalar_t,real_t>::~DistributedMatrixWrapper() {
    this->_data = NULL;
  }

  template<typename scalar_t,typename real_t> DistributedMatrix<scalar_t,real_t>::DistributedMatrix()
    : DistributedMatrix(-1, 0, 0, default_MB, default_NB) {
    p_row = p_col = np_rows = np_cols = -1; // make sure active() returns false if not constructed properly
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrix<scalar_t,real_t>::DistributedMatrix(const DistributedMatrix<scalar_t,real_t>& m)
    : np_rows(m.np_rows), np_cols(m.np_cols), p_row(m.p_row), p_col(m.p_col),
      loc_rows(m.loc_rows), loc_cols(m.loc_cols) {
    for (int i=0; i<9; i++) _desc[i] = m._desc[i];
    delete[] _data;
    _data = new scalar_t[loc_rows*loc_cols];
    std::copy(m._data, m._data+loc_rows*loc_cols, _data);
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrix<scalar_t,real_t>::DistributedMatrix(int blacs_ctxt, int M, int N)
    : DistributedMatrix(blacs_ctxt, M, N, default_MB, default_NB) {
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrix<scalar_t,real_t>::DistributedMatrix(int blacs_ctxt, int M, int N, scalar_t* A, int all_ctxt)
    : DistributedMatrix(blacs_ctxt, M, N, default_MB, default_NB) {
    set_sequential(A, all_ctxt);
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrix<scalar_t,real_t>::DistributedMatrix(int blacs_ctxt, int M, int N, int MB, int NB) {
    MB = std::max(1, MB);
    NB = std::max(1, NB);
    Cblacs_gridinfo(blacs_ctxt, &np_rows, &np_cols, &p_row, &p_col);
    if (p_row == -1 || p_col == -1) {
      loc_rows = loc_cols = 0;
      _data = NULL;
      descset(_desc, M, N, MB, NB, 0, 0, blacs_ctxt, std::max(loc_rows,1));
    } else {
      loc_rows = numroc(M, MB, p_row, 0, np_rows);
      loc_cols = numroc(N, NB, p_col, 0, np_cols);
      _data = new scalar_t[loc_rows*loc_cols];
      if (descinit(_desc, M, N, MB, NB, 0, 0, blacs_ctxt, std::max(loc_rows,1))) {
	std::cerr << "ERROR: Could not create DistributedMatrix descriptor!" << std::endl;
	exit(1);
      }
    }
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrix<scalar_t,real_t>::DistributedMatrix(int desc[9]) {
    std::copy(desc, desc+9, _desc);
    Cblacs_gridinfo(_desc[1], &np_rows, &np_cols, &p_row, &p_col);
    if (p_row == -1 || p_col == -1) {
      loc_rows = loc_cols = 0;
      _data = NULL;
    } else {
      loc_rows = numroc(_desc[2], _desc[4], p_row, _desc[6], np_rows);
      loc_cols = numroc(_desc[3], _desc[5], p_col, _desc[7], np_cols);
      assert(loc_rows==_desc[8]);
      if (loc_rows && loc_cols) _data = new scalar_t[loc_rows*loc_cols];
      else _data = NULL;
    }
  }

  template<typename scalar_t,typename real_t>
  DistributedMatrix<scalar_t,real_t>::~DistributedMatrix() {
    clear();
  }

  template<typename scalar_t,typename real_t> DistributedMatrix<scalar_t,real_t>&
  DistributedMatrix<scalar_t,real_t>::operator=(const DistributedMatrix<scalar_t,real_t>& m) {
    np_rows = m.np_rows;   np_cols = m.np_cols;
    p_row = m.p_row;       p_col = m.p_col;
    loc_rows = m.loc_rows; loc_cols = m.loc_cols;
    std::copy(m._desc, m._desc+9, _desc);
    delete[] _data;
    _data = new scalar_t[loc_rows*loc_cols];
    std::copy(m._data, m._data+loc_rows*loc_cols, _data);
    return *this;
  }

  template<typename scalar_t,typename real_t> DistributedMatrix<scalar_t,real_t>&
  DistributedMatrix<scalar_t,real_t>::operator=(DistributedMatrix<scalar_t,real_t>&& m) {
    np_rows = m.np_rows;   np_cols = m.np_cols;
    p_row = m.p_row;       p_col = m.p_col;
    loc_rows = m.loc_rows; loc_cols = m.loc_cols;
    std::copy(m._desc, m._desc+9, _desc);
    delete[] _data;
    _data = m._data;
    m._data = NULL;
    return *this;
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::clear() {
    delete[] _data; _data = NULL;
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::zero() {
    if (!active()) return;
    std::fill(_data, _data+loc_rows*loc_cols, scalar_t(0.));
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::fill(scalar_t v) {
    if (!active()) return;
    std::fill(_data, _data+loc_rows*loc_cols, v);
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::random() {
    if (!active()) return;
    auto rgen = make_random_generator<real_t>(p_row+np_rows*p_col);
    for (int c=0; c<local_cols(); ++c)
      for (int r=0; r<local_rows(); ++r)
	operator()(r,c) = rgen->get();
    STRUMPACK_FLOPS(static_cast<long long int>(local_rows())*local_cols()*rgen->flops_per_prng());
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::eye() {
    if (!active()) return;
    for (int c=0; c<local_cols(); ++c)
      for (int r=0; r<local_rows(); ++r)
	operator()(r,c) = (rowl2g(r) == coll2g(c)) ? scalar_t(1.) : scalar_t(0.);
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::print(std::string mat_name, int precision) {
    if (!active()) return;
    DistributedMatrix<scalar_t,real_t> B(ctxt(), rows(), cols(), rows(), cols());
    pgemr2d(rows(), cols(), data(), 1, 1, desc(), B.data(), 1, 1, B.desc(), ctxt());
    if (B.proc_row() == 0 && B.proc_col() == 0) {
      auto old_precision = std::cout.precision(precision);
      std::cout << mat_name << " = [" << std::endl;
      for (int r=0; r<B.local_rows(); r++) {
	for (int c=0; c<B.local_cols(); c++)
	  std::cout << B(r,c) << " ";
	std::cout << ";" << std::endl;
      }
      std::cout << "];" << std::endl;
      std::cout.precision(old_precision);
    }
  }

  // assume a is on the master in column major ordering, with ld == this->rows()
  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::set_sequential(scalar_t* a, int all_ctxt) {
    int a_desc[9];
    descset(a_desc, rows(), cols(), rows(), cols(), 0, 0, ctxt(), rows());
    pgemr2d(rows(), cols(), a, 1, 1, a_desc, this->data(), 1, 1, this->desc(), all_ctxt);
  }

  // assume a is on the master in column major ordering, with ld == this->rows()
  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::get_sequential(scalar_t* a, int all_ctxt) {
    int a_desc[9];
    descset(a_desc, rows(), cols(), rows(), cols(), 0, 0, ctxt(), rows());
    pgemr2d(rows(), cols(), this->data(), 1, 1, this->desc(), a, 1, 1, a_desc, all_ctxt);
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::LU(int* ipiv) {
    if (!active()) return;
    int info = pgetrf(rows(), cols(), data(), 1, 1, desc(), ipiv);
    if (info) {
      std::cerr << "ERROR: LU factorization of DistributedMatrix failed with info = " << info << std::endl;
      std::cerr << "This is probably due to a zero pivot!" << std::endl
		<< "Try running with MC64 pivoting (--sp_mc64job 5)" << std::endl;
      exit(1);
    }
#ifdef COUNT_FLOPS
    if (proc_row()==0 && proc_col()==0) {
      long long int M = rows();
      long long int N = cols();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_FLOPS(static_cast<long long int>(((M<N) ? (.5*double(M)*(double(M)*(double(N)-(1./3.)*double(M)-1.) + double(N)) + (2./3.) * double(M)) : (.5*double(N)*(double(N)*(double(M)-(1./3.)*double(N)-1.) + double(M)) + (2./3.) * double(N))) + ((M<N) ? (.5*double(M)*(double(M)*(double(N)-(1./3.)*double(M)) - double(N)) + (1./6.) * double(M)) : (.5*double(N)*(double(N)*(double(M)-(1./3.)*double(N)-1.) - double(M)) + (1./6.) * double(N)))));
      } else {
	STRUMPACK_FLOPS(4*static_cast<long long int>(((M<N) ? (.5*double(M)*(double(M)*(double(N)-(1./3.)*double(M)-1.) + double(N)) + (2./3.) * double(M)) : (.5*double(N)*(double(N)*(double(M)-(1./3.)*double(N)-1.) + double(M)) + (2./3.) * double(N))) + ((M<N) ? (.5*double(M)*(double(M)*(double(N)-(1./3.)*double(M)) - double(N)) + (1./6.) * double(M)) : (.5*double(N)*(double(N)*(double(M)-(1./3.)*double(N)-1.) - double(M)) + (1./6.) * double(N)))));
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t> inline void DistributedMatrix<scalar_t,real_t>::
  laswp(int* ipiv) {
    if (!active()) return;
    plaswp('F', 'R', cols(), data(), 1, 1, desc(), 1, rows(), ipiv);
#ifdef COUNT_FLOPS
    if (proc_row()==0 && proc_col()==0) {
      long long int N = cols();
      long long int k1 = 1;
      long long int k2 = rows();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_BYTES(4*static_cast<long long int>(2.*double(k2-k1)*double(N)));
      } else {
	STRUMPACK_BYTES(2*8*static_cast<long long int>(2.*double(k2-k1)*double(N)));
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::lapiv(int* ipiv, DistributedMatrix<scalar_t,real_t>& A) {
    if (!active()) return;
    int descip[9];
    descset(descip, A.rows() + A.MB()*A.proc_rows(), 1,
	    A.MB(), 1, 0, A.proc_col(), A.ctxt(),
	    A.MB() + numroc(A.rows(), A.MB(), A.proc_row(), 0, A.proc_rows()));
    plapiv('F', 'R', 'C', rows(), cols(), data(), 1, 1, desc(), ipiv, 1, 1, descip, NULL);
#ifdef COUNT_FLOPS
    if (proc_row()==0 && proc_col()==0) {
      long long int N = cols();
      long long int k1 = 1;
      long long int k2 = rows();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_BYTES(4*static_cast<long long int>(2.*double(k2-k1)*double(N)));
      } else {
	STRUMPACK_BYTES(2*8*static_cast<long long int>(2.*double(k2-k1)*double(N)));
      }
    }
#endif
  }

  // Solve a system of linear equations with B as right hand side.
  // assumption: the current matrix should have ben factored using LU.
  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::getrs(char trans, DistributedMatrix<scalar_t,real_t>& B, int* ipiv) {
    if (!active()) return;
    if (pgetrs(trans, B.rows(), B.cols(), data(), 1, 1, desc(), ipiv, B.data(), 1, 1, B.desc())) {
      std::cerr << "# ERROR: Failure in PGETRS :(" << std::endl; abort();
    }
#ifdef COUNT_FLOPS
    if (proc_row()==0 && proc_col()==0) {
      long long int N = B.rows();
      long long int nrhs = B.cols();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_FLOPS(static_cast<long long int>(2.*double(N)*double(N)*double(nrhs)));
      } else {
	STRUMPACK_FLOPS(4*static_cast<long long int>(2.*double(N)*double(N)*double(nrhs)));
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t> inline void
  gemm(char transA, char transB, scalar_t alpha, DistributedMatrix<scalar_t,real_t>& A, DistributedMatrix<scalar_t,real_t>& B, scalar_t beta, DistributedMatrix<scalar_t,real_t>& C) {
    if (!A.active()) return;
    pgemm(transA, transB, A.rows(), B.cols(), A.cols(), alpha, A.data(), 1, 1, A.desc(), B.data(), 1, 1, B.desc(), beta, C.data(), 1, 1, C.desc());
#ifdef COUNT_FLOPS
    if (A.proc_row()==0 && A.proc_col()==0) {
      long long int M = A.rows();
      long long int N = B.cols();
      long long int K = A.cols();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_FLOPS(static_cast<long long int>((alpha!=scalar_t(0))*double(M)*double(N)*(double(K)*2.-1.) + (alpha!=scalar_t(0) && beta!=scalar_t(0))*double(M)*double(N)
						   + (alpha!=scalar_t(0) && alpha!=scalar_t(1))*double(M)*double(N) + (beta!=scalar_t(0) && beta!=scalar_t(1))*double(M)*double(N)));
	STRUMPACK_BYTES(4*static_cast<long long int>(2.*double(M)*double(N)+double(M)*double(K)+double(K)*double(N)));
      } else {
	STRUMPACK_FLOPS(4*static_cast<long long int>((alpha!=scalar_t(0))*double(M)*double(N)*(double(K)*2.-1.) + (alpha!=scalar_t(0) && beta!=scalar_t(0))*double(M)*double(N)
						     + (alpha!=scalar_t(0) && alpha!=scalar_t(1))*double(M)*double(N) + (beta!=scalar_t(0) && beta!=scalar_t(1))*double(M)*double(N)));
	STRUMPACK_BYTES(2*8*static_cast<long long int>(2.*double(M)*double(N)+double(M)*double(K)+double(K)*double(N)));
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t> void
  trsm(char side, char uplo, char transA, char diag, scalar_t alpha,
       DistributedMatrix<scalar_t,real_t>& A, DistributedMatrix<scalar_t,real_t>& B) {
    if (!A.active()) return;
    ptrsm(side, uplo, transA, diag, B.rows(), B.cols(), alpha, A.data(), 1, 1, A.desc(), B.data(), 1, 1, B.desc());
#ifdef COUNT_FLOPS
    if (A.proc_row()==0 && A.proc_col()==0) {
      long long int M = B.rows();
      long long int N = B.cols();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_FLOPS(static_cast<long long int>((side=='L'||side=='l') ? (alpha!=scalar_t(0))*double(N)*double(M)*(double(M)+1.) + (alpha!=scalar_t(1) && alpha!=scalar_t(0))*double(N)*double(M) :
						   (alpha!=scalar_t(0))*double(M)*double(N)*(double(N)+1.) + (alpha!=scalar_t(1) && alpha!=scalar_t(0))*double(N)*double(M) ));
	STRUMPACK_BYTES(4*static_cast<long long int>(double(N)*double(N)*.5+2.*double(M)*double(N)));
      } else {
	STRUMPACK_FLOPS(4*static_cast<long long int>((side=='L'||side=='l') ? (alpha!=scalar_t(0))*double(N)*double(M)*(double(M)+1.) + (alpha!=scalar_t(1) && alpha!=scalar_t(0))*double(N)*double(M) :
						     (alpha!=scalar_t(0))*double(M)*double(N)*(double(N)+1.) + (alpha!=scalar_t(1) && alpha!=scalar_t(0))*double(N)*double(M) ));
	STRUMPACK_BYTES(2*8*static_cast<long long int>(double(N)*double(N)*.5+2.*double(M)*double(N)));
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t> void
  trsv(char uplo, char transA, char diag, DistributedMatrix<scalar_t,real_t>& A,
       DistributedMatrix<scalar_t,real_t>& B) {
    if (!A.active()) return;
    ptrsv(uplo, transA, diag, A.rows(), A.data(), 1, 1, A.desc(), B.data(), 1, 1, B.desc(), 1);
#ifdef COUNT_FLOPS
    if (A.proc_row()==0 && A.proc_col()==0) {
      long long int M = A.rows();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_FLOPS(static_cast<long long int>(double(M)*(double(M)+1.)));
	STRUMPACK_BYTES(4*static_cast<long long int>(double(M)*double(M)*.5+2.*double(M)));
      } else {
	STRUMPACK_FLOPS(4*static_cast<long long int>(double(M)*(double(M)+1.)));
	STRUMPACK_BYTES(2*8*static_cast<long long int>(double(M)*double(M)*.5+2.*double(M)));
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t> void
  gemv(char transA, scalar_t alpha, DistributedMatrix<scalar_t,real_t>& A,
       DistributedMatrix<scalar_t,real_t>& B, scalar_t beta, DistributedMatrix<scalar_t,real_t>& C) {
    if (!A.active()) return;
    assert(A.cols() == B.rows());
    assert(A.rows() == C.rows());
    assert(C.cols() == B.cols());
    pgemv(transA, A.rows(), A.cols(), alpha, A.data(), 1, 1, A.desc(),
	  B.data(), 1, 1, B.desc(), 1, beta, C.data(), 1, 1, C.desc(), 1);
#ifdef COUNT_FLOPS
    if (A.proc_row()==0 && A.proc_col()==0) {
      long long int M = A.rows();
      long long int N = A.cols();
      if (!is_complex<scalar_t>()) {
	STRUMPACK_FLOPS(static_cast<long long int>((alpha!=scalar_t(0))*double(M)*(double(N)*2.-1.) + (alpha!=scalar_t(1) && alpha!=scalar_t(0))*double(M) + (beta!=scalar_t(0) && beta!=scalar_t(1))*double(M) + (alpha!=scalar_t(0) && beta!=scalar_t(0))*double(M)));
	STRUMPACK_BYTES(4*static_cast<long long int>(2.*double(M)+double(M)*double(N)+double(N)));
      } else {
	STRUMPACK_FLOPS(4*static_cast<long long int>((alpha!=scalar_t(0))*double(M)*(double(N)*2.-1.) + (alpha!=scalar_t(1) && alpha!=scalar_t(0))*double(M) + (beta!=scalar_t(0) && beta!=scalar_t(1))*double(M) + (alpha!=scalar_t(0) && beta!=scalar_t(0))*double(M)));
	STRUMPACK_BYTES(2*8*static_cast<long long int>(2.*double(M)+double(M)*double(N)+double(N)));
      }
    }
#endif
  }

  /** sub( C ) := beta*sub( C ) + alpha*op( sub( A ) ) */
  template<typename scalar_t,typename real_t> void
  geadd(char trans, scalar_t alpha, DistributedMatrix<scalar_t,real_t>& A,
	scalar_t beta, DistributedMatrix<scalar_t,real_t>& C) {
    if (!A.active()) return;
    assert(A.rows() == C.rows());
    assert(A.cols() == C.cols());
    pgeadd(trans, A.rows(), A.cols(), alpha, A.data(), 1, 1, A.desc(), beta, C.data(), 1, 1, C.desc());
#ifdef COUNT_FLOPS
    // TODO flop/byte count
#endif
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extract_F11
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t sep_begin, integer_t dim_sep) {
    if (!active()) return;
    zero();
    const auto CB = local_col_blocks();
    const auto RB = local_row_blocks();
#pragma omp parallel for collapse(2)
    for (int cb=0; cb<CB; cb++) {
      for (int rb=0; rb<RB; rb++) {
	auto col = (proc_col()+cb*proc_cols())*NB();
	auto row = (proc_row()+rb*proc_rows())*MB();
	auto block = data() + cb*NB()*ld() + rb*MB();
	auto nr_cols = std::min(NB(), cols()-col);
	A->extract_F11_block(block, ld(), row+sep_begin, std::min(MB(),rows()-row), col+sep_begin, nr_cols);
      }
    }
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extract_F12
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t upd_row_begin,
   integer_t upd_col_begin, integer_t dim_upd, integer_t* upd) {
    if (!active()) return;
    zero();
    const auto CB = local_col_blocks();
    const auto RB = local_row_blocks();
#pragma omp parallel for collapse(2)
    for (int cb=0; cb<CB; cb++) {
      for (int rb=0; rb<RB; rb++) {
	auto col = (proc_col()+cb*proc_cols())*NB();
	auto row = (proc_row()+rb*proc_rows())*MB();
	auto block_upd = upd + (proc_col()+cb*proc_cols())*NB();
	auto nr_cols = std::min(NB(), cols()-col);
	auto block = data() + cb*NB()*ld() + rb*MB();
	A->extract_F12_block(block, ld(), row+upd_row_begin, std::min(MB(), rows()-row),
			     col+upd_col_begin, nr_cols, block_upd);
      }
    }
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extract_F21
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, integer_t upd_row_begin,
   integer_t upd_col_begin, integer_t dim_upd, integer_t* upd) {
    if (!active()) return;
    zero();
    const auto CB = local_col_blocks();
    const auto RB = local_row_blocks();
#pragma omp parallel for collapse(2)
    for (int cb=0; cb<CB; cb++) {
      for (int rb=0; rb<RB; rb++) {
	auto col = (proc_col()+cb*proc_cols())*NB();
	auto row = (proc_row()+rb*proc_rows())*MB();
	auto nr_cols = std::min(NB(), cols()-col);
	auto block = data() + cb*NB()*ld() + rb*MB();
	auto block_upd = upd + proc_row()*MB() + rb*proc_rows()*MB();
	A->extract_F21_block(block, ld(), row+upd_row_begin, std::min(MB(), rows()-row),
			     col+upd_col_begin, nr_cols, block_upd);
      }
    }
  }

  // this should be called on the child's F22, with the parent's front as argument
  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extend_add_copy_to_buffers
  (DistributedMatrix<scalar_t,real_t>& F11, DistributedMatrix<scalar_t,real_t>& F12,
   DistributedMatrix<scalar_t,real_t>& F21, DistributedMatrix<scalar_t,real_t>& F22,
   std::vector<std::vector<scalar_t>>& sbuf, FrontalMatrixDenseMPI<scalar_t,real_t,integer_t>* pa,
   std::vector<integer_t>& pa_ind) {
    auto pa_row = new int[local_rows()+local_cols()];
    auto pa_col = pa_row+local_rows();
    if (fixed()) {
      for (int r=0; r<local_rows(); r++) pa_row[r] = pa_ind[rowl2g_fixed(r)];
      for (int c=0; c<local_cols(); c++) pa_col[c] = pa_ind[coll2g_fixed(c)];
    } else {
      for (int r=0; r<local_rows(); r++) pa_row[r] = pa_ind[rowl2g(r)];
      for (int c=0; c<local_cols(); c++) pa_col[c] = pa_ind[coll2g(c)];
    }
    int r_upd, c_upd;
    for (r_upd=0; r_upd<local_rows(); r_upd++) if (pa_row[r_upd] >= pa->dim_sep) break;
    for (c_upd=0; c_upd<local_cols(); c_upd++) if (pa_col[c_upd] >= pa->dim_sep) break;
    if (fixed()) {
      for (int c=0; c<c_upd; c++) // F11
	for (int r=0; r<r_upd; r++)
	  sbuf[pa->find_rank_fixed(pa_row[r], pa_col[c], F11)].push_back(operator()(r,c));
      for (int c=c_upd; c<local_cols(); c++) // F12
	for (int r=0; r<r_upd; r++)
	  sbuf[pa->find_rank_fixed(pa_row[r], pa_col[c]-pa->dim_sep, F12)].push_back(operator()(r,c));
      for (int c=0; c<c_upd; c++) // F21
	for (int r=r_upd; r<local_rows(); r++)
	  sbuf[pa->find_rank_fixed(pa_row[r]-pa->dim_sep, pa_col[c], F21)].push_back(operator()(r,c));
      for (int c=c_upd; c<local_cols(); c++) // F22
	for (int r=r_upd; r<local_rows(); r++)
	  sbuf[pa->find_rank_fixed(pa_row[r]-pa->dim_sep, pa_col[c]-pa->dim_sep, F22)].push_back(operator()(r,c));
    } else {
      for (int c=0; c<c_upd; c++) // F11
	for (int r=0; r<r_upd; r++)
	  sbuf[pa->find_rank(pa_row[r], pa_col[c], F11)].push_back(operator()(r,c));
      for (int c=c_upd; c<local_cols(); c++) // F12
	for (int r=0; r<r_upd; r++)
	  sbuf[pa->find_rank(pa_row[r], pa_col[c]-pa->dim_sep, F12)].push_back(operator()(r,c));
      for (int c=0; c<c_upd; c++) // F21
	for (int r=r_upd; r<local_rows(); r++)
	  sbuf[pa->find_rank(pa_row[r]-pa->dim_sep, pa_col[c], F21)].push_back(operator()(r,c));
      for (int c=c_upd; c<local_cols(); c++) // F22
	for (int r=r_upd; r<local_rows(); r++)
	  sbuf[pa->find_rank(pa_row[r]-pa->dim_sep, pa_col[c]-pa->dim_sep, F22)].push_back(operator()(r,c));
    }
    delete[] pa_row;
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::skinny_extend_add_copy_to_buffers
  (DistributedMatrix<scalar_t,real_t>& S, std::vector<std::vector<scalar_t>>& sbuf,
   FrontalMatrixHSSMPI<scalar_t,real_t,integer_t>* pa, std::vector<integer_t>& pa_ind) {
    if (fixed()) {
      for (int c=0; c<local_cols(); c++) {
	auto gc = coll2g_fixed(c);
	for (int r=0; r<local_rows(); r++)
	  sbuf[pa->find_rank_fixed(pa_ind[rowl2g_fixed(r)], gc, S)].push_back(operator()(r,c));
      }
    } else {
      for (int c=0; c<local_cols(); c++) {
	auto gc = coll2g(c);
	for (int r=0; r<local_rows(); r++)
	  sbuf[pa->find_rank(pa_ind[rowl2g(r)], gc, S)].push_back(operator()(r,c));
      }
    }
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extend_add_column_copy_to_buffers
  (DistributedMatrix<scalar_t,real_t>& Bsep, DistributedMatrix<scalar_t,real_t>& Bupd,
   std::vector<std::vector<scalar_t>>& sbuf, FrontalMatrixMPI<scalar_t,real_t,integer_t>* pa,
   std::vector<integer_t>& pa_ind) {
    if (fixed()) {
      for (int r=0; r<local_rows(); r++) {
	auto pa_row = pa_ind[rowl2g_fixed(r)];
	if (pa_row < pa->dim_sep)
	  sbuf[pa->find_rank_fixed(pa_row, 0, Bsep)].push_back(operator()(r,0));
	else sbuf[pa->find_rank_fixed(pa_row-pa->dim_sep, 0, Bupd)].push_back(operator()(r,0));
      }
    } else {
      for (int r=0; r<local_rows(); r++) {
	auto pa_row = pa_ind[rowl2g(r)];
	if (pa_row < pa->dim_sep)
	  sbuf[pa->find_rank(pa_row, 0, Bsep)].push_back(operator()(r,0));
	else sbuf[pa->find_rank(pa_row-pa->dim_sep, 0, Bupd)].push_back(operator()(r,0));
      }
    }
  }

  /**
   * extract the elements pa_ind from Bupd and Bsep and put them into
   * the correct buffer to send to the child processes so they can be
   * put into child Bupd (this object)
   */
  template<typename _scalar_t,typename _real_t,typename integer_t> void extract_b_copy_to_buffers
  (DistributedMatrix<_scalar_t,_real_t>& Bsep, DistributedMatrix<_scalar_t,_real_t>& Bupd,
   std::vector<std::vector<_scalar_t>>& sbuf, std::function<int(integer_t)> ch_rank,
   std::vector<integer_t>& pa_ind, int ch_proc_rows) {
    integer_t pa_dim_sep = Bsep.rows();
    integer_t ch_dim_upd = pa_ind.size();
    integer_t ind_ptr = 0;
    for (int r=0; r<Bsep.local_rows(); r++) {
      auto gr = Bsep.rowl2g(r);
      while (ind_ptr < ch_dim_upd && pa_ind[ind_ptr] < gr) ind_ptr++;
      if (ind_ptr == ch_dim_upd) return;
      if (pa_ind[ind_ptr] >= pa_dim_sep) break;
      if (pa_ind[ind_ptr] != gr) continue;
      sbuf[ch_rank(ind_ptr)].push_back(Bsep(r,0));
    }
    for (int r=0; r<Bupd.local_rows(); r++) {
      auto gr = Bupd.rowl2g(r) + pa_dim_sep;
      while (ind_ptr < ch_dim_upd && pa_ind[ind_ptr] < gr) ind_ptr++;
      if (ind_ptr == ch_dim_upd) break;
      if (pa_ind[ind_ptr] != gr) continue;
      sbuf[ch_rank(ind_ptr)].push_back(Bupd(r,0));
    }
  }

  template<typename _scalar_t,typename _real_t,typename integer_t> void extend_add_copy_from_buffers
  (DistributedMatrix<_scalar_t,_real_t>& F11, DistributedMatrix<_scalar_t,_real_t>& F12,
   DistributedMatrix<_scalar_t,_real_t>& F21, DistributedMatrix<_scalar_t,_real_t>& F22,
   _scalar_t** pbuf, integer_t sep_begin, integer_t* pa_upd,
   integer_t* ch_upd, integer_t ch_dim_upd, std::function<int(integer_t,integer_t)> F22rank) {
    std::function<integer_t(integer_t)> sep_map = [&](integer_t i) { return i + sep_begin; };
    std::function<integer_t(integer_t)> upd_map = [&](integer_t i) { return pa_upd[i]; };
    F11.copy_from_buffer(pbuf, ch_upd, ch_dim_upd, F22rank, sep_map, sep_map);
    F12.copy_from_buffer(pbuf, ch_upd, ch_dim_upd, F22rank, sep_map, upd_map);
    F21.copy_from_buffer(pbuf, ch_upd, ch_dim_upd, F22rank, upd_map, sep_map);
    F22.copy_from_buffer(pbuf, ch_upd, ch_dim_upd, F22rank, upd_map, upd_map);
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::skinny_extend_add_copy_from_buffers
  (scalar_t** pbuf, integer_t sep_begin, integer_t dim_sep, integer_t* pa_upd,
   integer_t* ch_upd, integer_t ch_dim_upd, std::function<int(integer_t,integer_t)> chSrank) {
    if (!active()) return;
    auto rowf2g = [&](integer_t r) { return (r < dim_sep) ? r+sep_begin : pa_upd[r-dim_sep]; };
    if (fixed()) {
      for (int c=0; c<local_cols(); c++) {
	auto gc = coll2g_fixed(c);
	integer_t upd_r = 0;
	for (int r=0; r<local_rows(); r++) {
	  auto fgr = rowf2g(rowl2g_fixed(r));
	  while (upd_r < ch_dim_upd && ch_upd[upd_r] < fgr) upd_r++;
	  if (upd_r == ch_dim_upd) break;
	  if (ch_upd[upd_r] != fgr) continue;
	  auto src = chSrank(upd_r, gc);
	  operator()(r,c) += *(pbuf[src]);
	  pbuf[src]++;
	}
      }
    } else {
      for (int c=0; c<local_cols(); c++) {
	auto gc = coll2g(c);
	integer_t upd_r = 0;
	for (int r=0; r<local_rows(); r++) {
	  auto fgr = rowf2g(rowl2g(r));
	  while (upd_r < ch_dim_upd && ch_upd[upd_r] < fgr) upd_r++;
	  if (upd_r == ch_dim_upd) break;
	  if (ch_upd[upd_r] != fgr) continue;
	  auto src = chSrank(upd_r, gc);
	  operator()(r,c) += *(pbuf[src]);
	  pbuf[src]++;
	}
      }
    }
  }

  template<typename _scalar_t,typename _real_t,typename integer_t> void extend_add_column_copy_from_buffers
  (DistributedMatrix<_scalar_t,_real_t>& Bsep, DistributedMatrix<_scalar_t,_real_t>& Bupd,
   std::vector<std::vector<_scalar_t>>& buf, integer_t sep_begin, integer_t* pa_upd,
   integer_t* ch_upd, integer_t ch_dim_upd, std::function<int(integer_t,integer_t)> b_rank) {
    std::vector<_scalar_t*> pbuf(buf.size());
    for (size_t p=0; p<buf.size(); p++) pbuf[p] = buf[p].data();

    std::function<integer_t(integer_t)> sep_map = [&](integer_t i) { return i + sep_begin; };
    std::function<integer_t(integer_t)> upd_map = [&](integer_t i) { return pa_upd[i]; };
    Bsep.copy_column_from_buffer(pbuf, ch_upd, ch_dim_upd, b_rank, sep_map);
    Bupd.copy_column_from_buffer(pbuf, ch_upd, ch_dim_upd, b_rank, upd_map);
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extract_b_copy_from_buffers
  (std::vector<std::vector<scalar_t>>& buf, std::vector<integer_t>& pa_ind,
   std::function<int(integer_t)> src_rank) {
    std::vector<scalar_t*> pbuf(buf.size());
    for (size_t p=0; p<buf.size(); p++) pbuf[p] = buf[p].data();
    if (fixed()) {
      for (int r=0; r<local_rows(); r++) {
	auto rank = src_rank(pa_ind[rowl2g_fixed(r)]);
	operator()(r,0) = *(pbuf[rank]);
	pbuf[rank]++;
      }
    } else {
      for (int r=0; r<local_rows(); r++) {
	auto rank = src_rank(pa_ind[rowl2g(r)]);
	operator()(r,0) = *(pbuf[rank]);
	pbuf[rank]++;
      }
    }
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::copy_from_buffer
  (scalar_t** pbuf, integer_t* ch_upd, integer_t ch_dim_upd,
   std::function<int(integer_t,integer_t)> F22_child_rank,
   std::function<integer_t(integer_t)> front_r2g,
   std::function<integer_t(integer_t)> front_c2g) {
    if (!active()) return;
    // TODO precompute indices??
    //   list of local rows/col that have corresponding global index in front and corresponding child row
    // similar to compute_parent_indices and pass as input because this function is called 4 times
    if (fixed()) {
      integer_t upd_c = 0;
      for (int c=0; c<local_cols(); c++) {
    	auto fgc = front_c2g(coll2g_fixed(c));
    	while (upd_c < ch_dim_upd && ch_upd[upd_c] < fgc) upd_c++;
    	if (upd_c == ch_dim_upd) break;
    	if (ch_upd[upd_c] != fgc) continue;
    	integer_t upd_r = 0;
    	for (int r=0; r<local_rows(); r++) {
    	  auto fgr = front_r2g(rowl2g_fixed(r));
    	  while (upd_r < ch_dim_upd && ch_upd[upd_r] < fgr) upd_r++;
    	  if (upd_r == ch_dim_upd) break;
    	  if (ch_upd[upd_r] != fgr) continue;
    	  integer_t src = F22_child_rank(upd_r, upd_c);
    	  operator()(r,c) += *(pbuf[src]);
    	  pbuf[src]++;
    	}
      }
    } else {
      integer_t upd_c = 0;
      for (int c=0; c<local_cols(); c++) {
    	auto fgc = front_c2g(coll2g(c));
    	while (upd_c < ch_dim_upd && ch_upd[upd_c] < fgc) upd_c++;
    	if (upd_c == ch_dim_upd) break;
    	if (ch_upd[upd_c] != fgc) continue;
    	integer_t upd_r = 0;
    	for (int r=0; r<local_rows(); r++) {
    	  auto fgr = front_r2g(rowl2g(r));
    	  while (upd_r < ch_dim_upd && ch_upd[upd_r] < fgr) upd_r++;
    	  if (upd_r == ch_dim_upd) break;
    	  if (ch_upd[upd_r] != fgr) continue;
    	  integer_t src = F22_child_rank(upd_r, upd_c);
    	  operator()(r,c) += *(pbuf[src]);
    	  pbuf[src]++;
    	}
      }
    }
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::copy_column_from_buffer
  (std::vector<scalar_t*>& pbuf, integer_t* ch_upd, integer_t ch_dim_upd,
   std::function<int(integer_t,integer_t)> b_child_rank,
   std::function<integer_t(integer_t)> front_to_global) {
    if (!active()) return;
    integer_t upd_r = 0;
    for (int r=0; r<local_rows(); r++) {
      auto fgr = front_to_global(rowl2g(r));
      while (upd_r < ch_dim_upd && ch_upd[upd_r] < fgr) upd_r++;
      if (upd_r == ch_dim_upd) break;
      if (ch_upd[upd_r] != fgr) continue;
      integer_t src = b_child_rank(upd_r, 0);
      operator()(r,0) += *(pbuf[src]);
      pbuf[src]++;
    }
  }

  /**
   * Extract elements [rows,cols] from this matrix. Results should go
   * to mat[row_pos,col_pos]. This function copies the elements
   * [rows,cols] from this matrix to the send buffers, so that they
   * can be unpacked and put into mat later by a call to
   * extract_copy_from_buffers.
   *
   * Here we can ask the rank directly from mat, because mat is
   * defined on the same context as this matrix.
   */
  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extract_copy_to_buffers
  (std::vector<integer_t>& rows, std::vector<integer_t>& cols,
   std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
   DistributedMatrix<scalar_t,real_t>& mat, std::vector<std::vector<scalar_t>>& sbuf) {
    if (!active()) return;
    if (fixed()) {
      for (size_t c=0; c<cols.size(); c++) {
	auto gc = cols[c];
	if (colg2p_fixed(gc) != proc_col()) continue;
	auto lc = colg2l_fixed(gc);
	for (size_t r=0; r<rows.size(); r++) {
	  auto gr = rows[r];
	  if (rowg2p_fixed(gr) == proc_row()) {
	    auto dest = mat.rank_fixed(row_pos[r], col_pos[c]);
	    sbuf[dest].push_back(operator()(rowg2l_fixed(gr),lc));
	  }
	}
      }
    } else {
      for (size_t c=0; c<cols.size(); c++) {
	auto gc = cols[c];
	if (colg2p(gc) != proc_col()) continue;
	auto lc = colg2l(gc);
	for (size_t r=0; r<rows.size(); r++) {
	  auto gr = rows[r];
	  if (rowg2p(gr) == proc_row()) {
	    auto dest = mat.rank(row_pos[r], col_pos[c]);
	    sbuf[dest].push_back(operator()(rowg2l(gr),lc));
	  }
	}
      }
    }
  }

  /** See description of extract_copy_to_buffers */
  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extract_copy_from_buffers
  (std::vector<integer_t>& rows, std::vector<integer_t>& cols,
   std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
   DistributedMatrix<scalar_t,real_t>& mat, scalar_t** pbuf) {
    if (!active()) return;
    if (fixed()) {
      for (size_t c=0; c<col_pos.size(); c++) {
	auto gc = col_pos[c];
	if (colg2p_fixed(gc) != proc_col()) continue;
	auto lc = colg2l_fixed(gc);
	for (size_t r=0; r<row_pos.size(); r++) {
	  auto gr = row_pos[r];
	  if (rowg2p_fixed(gr) == proc_row()) {
	    auto src = mat.rank_fixed(rows[r], cols[c]);
	    operator()(rowg2l_fixed(gr),lc) += *(pbuf[src]);
	    pbuf[src]++;
	  }
	}
      }
    } else {
      for (size_t c=0; c<col_pos.size(); c++) {
	auto gc = col_pos[c];
	if (colg2p(gc) != proc_col()) continue;
	auto lc = colg2l(gc);
	for (size_t r=0; r<row_pos.size(); r++) {
	  auto gr = row_pos[r];
	  if (rowg2p(gr) == proc_row()) {
	    auto src = mat.rank(rows[r], cols[c]);
	    operator()(rowg2l(gr),lc) += *(pbuf[src]);
	    pbuf[src]++;
	  }
	}
      }
    }
  }

  /**
   * Extend this smaller matrix to the larger matrix mat. The elements
   * go to [row_pos:col_pos] in mat.
   *
   * Here we can ask the rank directly from mat, because mat is
   * defined on the same context as this matrix.
   */
  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extend_copy_to_buffers
  (std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
   DistributedMatrix<scalar_t,real_t>& mat, std::vector<std::vector<scalar_t>>& sbuf) {
    if (!active()) return;
    if (fixed()) {
      for (integer_t c=0; c<local_cols(); c++) {
	auto pcol = col_pos[coll2g_fixed(c)];
	for (integer_t r=0; r<local_rows(); r++)
	  sbuf[mat.rank_fixed(row_pos[rowl2g_fixed(r)], pcol)].push_back(operator()(r,c));
      }
    } else {
      for (integer_t c=0; c<local_cols(); c++) {
	auto pcol = col_pos[coll2g(c)];
	for (integer_t r=0; r<local_rows(); r++)
	  sbuf[mat.rank(row_pos[rowl2g(r)], pcol)].push_back(operator()(r,c));
      }
    }
  }

  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::extend_copy_from_buffers
  (std::vector<integer_t>& row_pos, std::vector<integer_t>& col_pos,
   DistributedMatrix<scalar_t,real_t>& mat, scalar_t** pbuf) {
    if (!active()) return;
    if (fixed()) {
      for (size_t c=0; c<col_pos.size(); c++) {
	auto gc = col_pos[c];
	if (colg2p_fixed(gc) != proc_col()) continue;
	auto lc = colg2l_fixed(gc);
	for (size_t r=0; r<row_pos.size(); r++) {
	  auto gr = row_pos[r];
	  if (rowg2p_fixed(gr) == proc_row()) {
	    auto src = mat.rank_fixed(r, c);
	    operator()(rowg2l_fixed(gr),lc) += *(pbuf[src]);
	    pbuf[src]++;
	  }
	}
      }
    } else {
      for (size_t c=0; c<col_pos.size(); c++) {
	auto gc = col_pos[c];
	if (colg2p(gc) != proc_col()) continue;
	auto lc = colg2l(gc);
	for (size_t r=0; r<row_pos.size(); r++) {
	  auto gr = row_pos[r];
	  if (rowg2p(gr) == proc_row()) {
	    auto src = mat.rank(r, c);
	    operator()(rowg2l(gr),lc) += *(pbuf[src]);
	    pbuf[src]++;
	  }
	}
      }
    }
  }

  /**
   * fill the matrix with random numbers such that corresponding rows in
   * child, parent and sibling have the same random numbers, regardless
   * of number of processes in the context.
   */
  template<typename scalar_t,typename real_t> template<typename integer_t> void
  DistributedMatrix<scalar_t,real_t>::fill_random
  (std::function<integer_t(integer_t)> global_row, integer_t dimA) {
    auto rgen = make_random_generator<real_t>();
    if (!active()) return;
    if (fixed()) {
      for (int cb=0; cb<local_col_blocks(); cb++) {
	auto gc = coll2g_fixed(cb*NB());
	for (int r=0; r<local_rows(); r++) {
	  auto gr = global_row(rowl2g_fixed(r));
	  rgen->seed(gr+1 + gc*dimA);
	  for (int c=cb*NB(); c<std::min((cb+1)*NB(),local_cols()); c++)
	    operator()(r,c) = rgen->get();
	}
      }
    } else {
      for (int cb=0; cb<local_col_blocks(); cb++) {
	auto gc = coll2g(cb*NB());
	for (int r=0; r<local_rows(); r++) {
	  auto gr = global_row(rowl2g(r));
	  rgen->seed(gr+1 + gc*dimA);
	  for (int c=cb*NB(); c<std::min((cb+1)*NB(),local_cols()); c++)
	    operator()(r,c) = rgen->get();
	}
      }
    }
    STRUMPACK_FLOPS(static_cast<long long int>(local_rows())*local_cols()*rgen->flops_per_prng());
  }

#if defined(WRITE_ROOT)
  template<typename scalar_t,typename real_t> inline void
  DistributedMatrix<scalar_t,real_t>::MPI_binary_write() {
    if (active()) {
      MPI_Comm comm = Cblacs2sys_handle(ctxt());
      long long* allbufsize = new long long[np_rows*np_cols];
      long long bufsize = loc_rows*loc_cols;
      MPI_Allgather(&bufsize, 1, mpi_type<long long>(), allbufsize, 1, mpi_type<long long>(), comm);

      long long disp = 0;
      for (int i=0; i<np_rows*p_col+p_row; i++) disp += allbufsize[i];
      delete[] allbufsize;
      disp *= sizeof(scalar_t); // Offset in bytes

      MPI_File fp;
      MPI_File_open(comm, "front", MPI_MODE_WRONLY+MPI_MODE_CREATE, MPI_INFO_NULL, &fp);
      MPI_File_set_view(fp, disp, mpi_type<scalar_t>(), mpi_type<scalar_t>(), "native", MPI_INFO_NULL);
      MPI_File_write(fp, _data, bufsize, mpi_type<scalar_t>(), MPI_STATUS_IGNORE);
      MPI_Barrier(comm);
      MPI_File_close(&fp);

      if (!p_row && !p_col) {
	FILE *fpinfo = fopen("front_info","w");
	// M, N, MB, NB, grid rows, grid cols, type size
	fprintf(fpinfo,"%d %d %d %d %d %d %d\n", rows(), cols(), MB(), NB(),
		proc_rows(), proc_cols(), sizeof(scalar_t));
	fclose(fpinfo);
      }
    }
  }
#endif

} // end namespace strumpack

#endif // DISTRIBUTED_MATRIX_HPP
