/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef PT_SCOTCH_REORDERING_HPP
#define PT_SCOTCH_REORDERING_HPP

#include <ptscotch.h>
#include <cassert>
#include <type_traits>
#include "ScotchReordering.hpp"
#include "CSRMatrixMPI.hpp"
#include "ETree.hpp"
#include "Redistribute.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  class PTScotchReordering : public MatrixReorderingMPI<scalar_t,real_t,integer_t> {
  public:
    PTScotchReordering(MPI_Comm c) : MatrixReorderingMPI<scalar_t,real_t,integer_t>(c) {}
    ~PTScotchReordering() {}
    std::string get_name() { return std::string("PTScotch"); }
    int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A,
			  bool diag, int stratnum, int stratpar, bool verbose,
			  int nx, int ny, int nz);
  protected:
    virtual std::string get_strategy(int stratnum, int stratpar);
    integer_t sep_tree_from_ptscotch_nd_tree(std::vector<SCOTCH_Num>& ptscotch_tree,
					     std::vector<SCOTCH_Num>& ptscotch_sizes);
  };

  template<typename scalar_t,typename real_t,typename integer_t> std::string
  PTScotchReordering<scalar_t,real_t,integer_t>::get_strategy(int stratnum, int stratpar) {
    std::stringstream strategy_string;
    switch (stratnum) {
    case 1:  // based on a string from MUMPS, only performs nested-dissection for log(P) levels
      strategy_string <<
	"n{"           // nested dissection
	"ole=s,"    // "simple" parallel ordering strategy for each distributed leaf of parallel separator tree
	"ose=s,"    // "simple" parallel ordering strategy for each distributed separator of the separator tree
	"osq=s,"    // "simple" sequential ordering on centralized subgraphs, after n-d has gone far enough that subgraph is on 1 process
	"sep=m{"    // use parallel vertex multi-level method to find new separators
	"asc=b{width=3,strat=q{strat=f}},"   // use band method to refine distr vert seps after uncoarsening:
	"low=q{strat=h},"                    // use multi-sequential method to compute sep of coarsest graph
	"vert=1000,"                         // threshold minimum size under which graph is no longer coarsened
	"dvert=100,"                         // avg number of vert for proc under which the folding process is performed during coarsening
	"dlevl=0"                            // minimum level after which duplication is allowed in folding process
	"}"
	"}"; break;
    case 2:

    default:  // perform nested dissection all the way down
      // TODO stop when separator smaller than stratpar!
      strategy_string <<
	"n{"           // nested dissection
	"ole=s,"    // "simple" parallel ordering strategy for each distributed leaf of parallel separator tree
	"ose=s,"    // "simple" parallel ordering strategy for each distributed separator of the separator tree
	"osq=n{ole=s,ose=s,sep=g},"    // nested-dissection ordering on centralized subgraphs with Gibbs-Poole-Stockmeyer to find seps
	"sep=m{"                       // use parallel vertex multi-level method to find new separators
	"asc=b{width=3,strat=q{strat=f}}," // use band method to refine distr vert seps after uncoarsening:
	"low=q{strat=h},"                  // use multi-sequential method to compute sep of coarsest graph
	"vert=1000,"                       // threshold minimum size under which graph is no longer coarsened
	"dvert=100,"                       // avg number of vert for proc under which the folding process is performed during coarsening
	"dlevl=0"                          // minimum level after which duplication is allowed in folding process
	"}"
	"}"; break;
    }
    return strategy_string.str();
  }

  /**
   * Take a ptscotch tree for the distributed part of nested-dissection
   * and make a regular binary tree.
   *
   * The ptscotch_tree layout:
   *   - ptscotch_tree[i] is parent of node i
   *   - the nodes with count 3 are nested-dissection nodes
   *     children are: (left_child, right_child, separator)
   *   - nodes with 2 children are nested dissection nodes with empty separator
   *   - ptscotch_sizes denote size of part plus all descendants
   *
   * return: number of separators
   *
   * TODO can there be multiple roots?
   * TODO level info
   */
  template<typename scalar_t,typename real_t,typename integer_t> integer_t
  PTScotchReordering<scalar_t,real_t,integer_t>::sep_tree_from_ptscotch_nd_tree
  (std::vector<SCOTCH_Num>& ptscotch_tree, std::vector<SCOTCH_Num>& ptscotch_sizes) {
    assert(std::count(ptscotch_tree.begin(), ptscotch_tree.end(), -1) == 1);
    std::vector<integer_t> count(ptscotch_tree.size(), 0);
    std::vector<integer_t> ptscotch_lchild(ptscotch_tree.size(), integer_t(-1));
    std::vector<integer_t> ptscotch_rchild(ptscotch_tree.size(), integer_t(-1));
    integer_t nr_sep = 0;
    for (size_t i=0; i<ptscotch_tree.size(); i++) {
      integer_t p = ptscotch_tree[i];
      nr_sep++;
      if (p != -1) {
	count[p]++;
	switch (count[p]) {
	case 1: ptscotch_lchild[p] = i; break;
	case 2: ptscotch_rchild[p] = i; break;
	case 3: nr_sep--; break;
	  // TODO what about case 4??
	}
      }
    }
    integer_t root_id = std::distance
      (ptscotch_tree.begin(), std::find(ptscotch_tree.begin(), ptscotch_tree.end(), integer_t(-1)));
    this->dist_sizes.resize(nr_sep);
    this->dist_sep_ptr.resize(nr_sep+1);
    this->dist_lchild.assign(nr_sep, integer_t(-1));
    this->dist_rchild.assign(nr_sep, integer_t(-1));
    this->dist_parent.assign(nr_sep, integer_t(-1));
    this->dist_level.resize(nr_sep);
    std::function<void(integer_t,integer_t&,integer_t)> f = [&](integer_t i, integer_t& pid, integer_t level) {
      if (ptscotch_lchild[i] != -1 && ptscotch_rchild[i] != -1) {
	f(ptscotch_lchild[i], pid, level+1);
	integer_t left_root_id = pid - 1;
	f(ptscotch_rchild[i], pid, level+1);
	this->dist_rchild[pid] = pid - 1;
	this->dist_parent[this->dist_rchild[pid]] = pid;
	this->dist_lchild[pid] = left_root_id;
	this->dist_parent[this->dist_lchild[pid]] = pid;
      }
      this->dist_level[pid] = level;
      this->dist_sizes[pid] = ptscotch_sizes[i];
      if (ptscotch_lchild[i] != -1 && ptscotch_rchild[i] != -1)
	this->dist_sizes[pid] -= ptscotch_sizes[ptscotch_lchild[i]] + ptscotch_sizes[ptscotch_rchild[i]];
      this->dist_sep_ptr[pid+1] = this->dist_sep_ptr[pid] + this->dist_sizes[pid];
      pid++;
    };
    nr_sep = 0;
    f(root_id, nr_sep, 0);
    return nr_sep;
  }

  template<typename integer_t> inline int WRAPPER_SCOTCH_dgraphOrderPerm
  (integer_t n, SCOTCH_Dgraph* graph, SCOTCH_Dordering* ordeptr, integer_t* order) {
    SCOTCH_Num* permloctab = new SCOTCH_Num[n];
    int ierr = SCOTCH_dgraphOrderPerm(graph, ordeptr, permloctab);
    if (ierr) return ierr;
    std::copy(permloctab, permloctab+n, order);
    return 0;
  }
  template<> inline int WRAPPER_SCOTCH_dgraphOrderPerm
  (SCOTCH_Num n, SCOTCH_Dgraph* graph, SCOTCH_Dordering* ordeptr, SCOTCH_Num* order) {
    return SCOTCH_dgraphOrderPerm(graph, ordeptr, order);
  }

  template<typename scalar_t,typename real_t,typename integer_t> int
  PTScotchReordering<scalar_t,real_t,integer_t>::nested_dissection
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
   int stratnum, int stratpar, bool verbose, int, int, int) {
    CSRMatrixMPI<scalar_t,real_t,integer_t>* A_mpi = dynamic_cast<CSRMatrixMPI<scalar_t,real_t,integer_t>*>(A);
    if (A_mpi == NULL) {
      std::cerr << "PTScotch reordering only works on CSRMatrixMPI (for now)!" << std::endl;
      return 1;
    }
    auto local_rows = A_mpi->local_rows();
    auto n = A_mpi->size();
    auto ptr = A_mpi->get_ptr();
    auto ind = A_mpi->get_ind();

    std::vector<SCOTCH_Num> ptr_nodiag(local_rows+1);
    std::vector<SCOTCH_Num> ind_nodiag(ptr[local_rows]-ptr[0]);
    integer_t nnz_nodiag = 0;
    ptr_nodiag[0] = 0;
    for (integer_t i=0; i<local_rows; i++) { // remove diagonal elements
      for (integer_t j=ptr[i]-ptr[0]; j<ptr[i+1]-ptr[0]; j++)
	if (ind[j]!=i+A_mpi->begin_row()) ind_nodiag[nnz_nodiag++] = ind[j];
      ptr_nodiag[i+1] = nnz_nodiag;
    } // ptr_nodiag will now point locally, ind_nodiag still has global column/row indices

    SCOTCH_Dgraph graph;
    int ierr = SCOTCH_dgraphInit(&graph, this->comm); if (ierr) return ierr;
    ierr = SCOTCH_dgraphBuild(&graph, 0, local_rows, local_rows, ptr_nodiag.data(), ptr_nodiag.data()+1,
			      NULL, NULL, nnz_nodiag, nnz_nodiag, ind_nodiag.data(), NULL, NULL);
    if (ierr) return ierr;
    assert(SCOTCH_dgraphCheck(&graph) == 0);
    SCOTCH_Strat strategy;
    ierr = SCOTCH_stratInit(&strategy); if (ierr) return ierr;
    ierr = SCOTCH_stratDgraphOrder(&strategy, this->get_strategy(stratnum, stratpar).c_str()); if (ierr) return ierr;
    SCOTCH_Dordering ordeptr;
    ierr = SCOTCH_dgraphOrderInit(&graph, &ordeptr); if (ierr) return ierr;
    ierr = SCOTCH_dgraphOrderCompute(&graph, &ordeptr, &strategy); if (ierr) return ierr;

    // get info about the distributed levels of nested-dissection
    this->dist_nr_sep = SCOTCH_dgraphOrderCblkDist(&graph, &ordeptr); if (ierr) return ierr;
    std::vector<SCOTCH_Num> ptscotch_dist_tree(this->dist_nr_sep);
    std::vector<SCOTCH_Num> ptscotch_dist_sizes(this->dist_nr_sep);
    ierr = SCOTCH_dgraphOrderTreeDist(&graph, &ordeptr, ptscotch_dist_tree.data(),
				      ptscotch_dist_sizes.data()); if (ierr) return ierr;
    // build separator tree for distributed nested dissection part
    this->dist_nr_sep = sep_tree_from_ptscotch_nd_tree(ptscotch_dist_tree, ptscotch_dist_sizes);

    integer_t* local_order = new integer_t[local_rows];
    ierr = WRAPPER_SCOTCH_dgraphOrderPerm(local_rows, &graph, &ordeptr, local_order); if (ierr) return ierr;

    this->allgather_local_order(A_mpi, local_order);
    delete[] local_order;

    this->get_local_graphs(A_mpi);

    this->build_sep_tree_from_order(A_mpi);

    SCOTCH_dgraphOrderExit(&graph, &ordeptr);
    SCOTCH_dgraphExit(&graph);
    SCOTCH_stratExit(&strategy);
    this->nested_dissection_print(n, A->nnz(), stratnum, stratpar, verbose);
    return 0;
  }

} // end namespace strumpack

#endif
