/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */
#ifndef SCOTCH_REORDERING_HPP
#define SCOTCH_REORDERING_HPP

#include "MatrixReordering.hpp"
#include "ETree.hpp"

namespace strumpack {

  template<typename scalar_t,typename real_t,typename integer_t>
  class ScotchReordering : public MatrixReordering<scalar_t,real_t,integer_t> {
  public:
    ScotchReordering() : MatrixReordering<scalar_t,real_t,integer_t>() {}
    ~ScotchReordering() {}
    std::string get_name() { return std::string("Scotch"); }
    virtual int nested_dissection(CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag,
				  int stratnum, int stratpar, bool verbose,
				  int nx, int ny, int nz);
  protected:
    virtual std::string get_strategy(int stratnum, int stratpar);
  private:
    void sep_tree_from_scotch_tree(std::vector<SCOTCH_Num>& scotch_tree,
				   std::vector<SCOTCH_Num>& scotch_sizes);
  };

  template<typename scalar_t,typename real_t,typename integer_t> std::string
  ScotchReordering<scalar_t,real_t,integer_t>::get_strategy(int stratnum, int stratpar) {
    std::stringstream strategy_string;
    switch (stratnum) {
    case 1: // nested dissection with STRATPAR levels
      strategy_string << "n{ole=s,ose=s,sep=(/levl>" << int(stratpar-1) << "?z:"
	"(m{asc=b{bnd=f{move=200,pass=1000,bal=0.9},org=(|h{pass=10})f{move=200,pass=1000,bal=0.9},width=3},"
	"low=h{pass=10},type=h,vert=100,rat=0.7});)}"; break;
    case 2: // nested dissection with separators <= STRATPAR
      strategy_string << "n{ole=s,ose=s,sep=(/vert<" << int(stratpar) << "?z:"
	"(m{asc=b{bnd=f{move=200,pass=1000,bal=0.9},org=(|h{pass=10})"
	"f{move=200,pass=1000,bal=0.9},width=3},"
	"low=h{pass=10},type=h,vert=100,rat=0.7});)}"; break;
    default: // Pure nested dissection
      strategy_string << "n{ole=s,ose=s,sep=m{asc=b{bnd=f{move=200,pass=1000,bal=0.9},"
	"org=(|h{pass=10})f{move=200,pass=1000,bal=0.9},width=3},low=h{pass=10},type=h,vert=100,rat=0.7}}";
    }
    return strategy_string.str();
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  ScotchReordering<scalar_t,real_t,integer_t>::sep_tree_from_scotch_tree
  (std::vector<SCOTCH_Num>& scotch_tree, std::vector<SCOTCH_Num>& scotch_sizes) {
    this->sep_ptr.assign(scotch_sizes.begin(), scotch_sizes.end());
    this->parent.assign(scotch_tree.begin(), scotch_tree.end());
    this->left_child.resize(this->nbsep, integer_t(-1));
    this->right_child.resize(this->nbsep, integer_t(-1));
    std::vector<integer_t> count(this->nbsep, 0);
    for (integer_t i=0; i<this->nbsep; i++) {
      integer_t p = this->parent[i];
      if (p != -1) {
	count[p]++;
	switch (count[p]) {
	case 1: this->left_child[p] = i; break; // node i is the first child of node p
	case 2: this->right_child[p] = i; break; // node i is the second child of node p
	case 3: // node i is the third child of node p
	  // TODO deal with this, see also PTScotchReordering!!
	  std::cerr << "ERROR: scotch tree has subleafs in the separators/leafs, should handle this!!" << std::endl;
	  break;
	}
      }
    }
    this->level.resize(this->nbsep, 0);
    std::function<void(integer_t,integer_t)> f = [&](integer_t i, integer_t l) {
      if (this->left_child[i] != -1) f(this->left_child[i], l+1);
      if (this->right_child[i] != -1) f(this->right_child[i], l+1);
      this->level[i] = l;
    };
    f(this->nbsep-1, 0);
    integer_t max_level = *std::max_element(this->level.begin(), this->level.end());
    std::for_each(this->level.begin(), this->level.end(), [&](integer_t& l){ l=max_level-l+1; });
  }

  // base template, integer_t != SCOTCH_Num, so need to copy everything
  template<typename integer_t> inline int WRAPPER_SCOTCH_graphOrder
  (integer_t n, SCOTCH_Graph* graph, SCOTCH_Strat* strat, integer_t* permtab, integer_t* peritab,
   integer_t* cblkptr, SCOTCH_Num* rangtab, SCOTCH_Num* treetab) {
    SCOTCH_Num  scotch_cblkptr;
    SCOTCH_Num* scotch_permtab = new SCOTCH_Num[2*n];
    SCOTCH_Num* scotch_peritab = scotch_permtab + n;
    int ierr = SCOTCH_graphOrder(graph, strat, scotch_permtab, scotch_peritab,
				 &scotch_cblkptr, rangtab, treetab);
    if (ierr) return ierr;
    std::copy(scotch_permtab, scotch_permtab+n, permtab);
    std::copy(scotch_peritab, scotch_peritab+n, peritab);
    delete[] scotch_permtab;
    *cblkptr = scotch_cblkptr;
    return 0;
  }
  // if integer_t == SCOTCH_Num, no need to copy
  template<> inline int WRAPPER_SCOTCH_graphOrder
  (SCOTCH_Num n, SCOTCH_Graph* graph, SCOTCH_Strat* strat, SCOTCH_Num* permtab, SCOTCH_Num* peritab,
   SCOTCH_Num* cblkptr, SCOTCH_Num* rangtab, SCOTCH_Num* treetab) {
    return SCOTCH_graphOrder(graph, strat, permtab, peritab, cblkptr, rangtab, treetab);
  }

  template<typename scalar_t,typename real_t,typename integer_t> int
  ScotchReordering<scalar_t,real_t,integer_t>::nested_dissection
  (CompressedSparseMatrix<scalar_t,real_t,integer_t>* A, bool diag, int stratnum, int stratpar,
   bool verbose, int, int, int) {
    integer_t n = A->size();
    integer_t* ptr = A->get_ptr();
    integer_t* ind = A->get_ind();

    SCOTCH_Graph graph;
    SCOTCH_graphInit(&graph);
    auto ptr_nodiag = new SCOTCH_Num[n+1 + ptr[n]-ptr[0]];
    auto ind_nodiag = ptr_nodiag + n+1;
    integer_t nnz_nodiag = 0;
    ptr_nodiag[0] = 0;
    for (integer_t i=0; i<n; i++) { // remove diagonal elements
      for (integer_t j=ptr[i]; j<ptr[i+1]; j++)
	if (ind[j]!=i) ind_nodiag[nnz_nodiag++] = ind[j];
      ptr_nodiag[i+1] = nnz_nodiag;
    }
    if (nnz_nodiag==0) std::cerr << "# WARNING: matrix seems to be diagonal!" << std::endl;
    int ierr = SCOTCH_graphBuild(&graph, 0, static_cast<SCOTCH_Num>(n), ptr_nodiag,
				 NULL, NULL, NULL, ptr_nodiag[n], ind_nodiag, NULL);
    if (ierr) return ierr;
    ierr = SCOTCH_graphCheck(&graph); if (ierr) return ierr;

    SCOTCH_Strat strategy;
    ierr = SCOTCH_stratInit(&strategy); if (ierr) return ierr;
    ierr = SCOTCH_stratGraphOrder(&strategy, get_strategy(stratnum, stratpar).c_str()); if (ierr) return ierr;

    std::vector<SCOTCH_Num> scotch_sizes(n+1);
    std::vector<SCOTCH_Num> scotch_tree(n);
    this->order.resize(n);
    this->iorder.resize(n);
    ierr = WRAPPER_SCOTCH_graphOrder<integer_t>
      (n, &graph, &strategy, this->order.data(), this->iorder.data(),
       &this->nbsep, scotch_sizes.data(), scotch_tree.data());
    if (ierr) return ierr;
    delete[] ptr_nodiag;

    SCOTCH_graphExit(&graph);
    SCOTCH_stratExit(&strategy);

    scotch_tree.resize(this->nbsep);
    scotch_sizes.resize(this->nbsep+1);
    if (std::count(scotch_tree.begin(), scotch_tree.end(), -1) > 1) {
      // the graph is disconnected, add one root to connect all
      SCOTCH_Num root_id = *std::max_element(scotch_tree.begin(), scotch_tree.end()) + 1;
      scotch_tree.push_back(root_id);
      scotch_sizes.push_back(scotch_sizes.back());
      std::replace(scotch_tree.begin(), scotch_tree.end(), SCOTCH_Num(-1), root_id);
      this->nbsep++;
    }
    sep_tree_from_scotch_tree(scotch_tree, scotch_sizes);

    this->nested_dissection_print(n, ptr[n]-ptr[0], stratnum, stratpar, verbose);
    return 0;
  }

} // end namespace strumpack

#endif
