/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 */
/*! \file StrumpackSparseSolverMPIDist.hpp
 * \brief Contains the definition of fully distributed sparse solver class.
 */
#ifndef STRUMPACK_SPARSE_SOLVER_MPI_DIST_H
#define STRUMPACK_SPARSE_SOLVER_MPI_DIST_H

#include "StrumpackSparseSolverMPI.hpp"
#include "EliminationTreeMPI.hpp"
#include "EliminationTreeMPIDist.hpp"
#include "MatrixReorderingMPI.hpp"
#include "ParMetisReordering.hpp"
#include "PTScotchReordering.hpp"
#include "GeometricReorderingMPI.hpp"
#include "GMResMPI.hpp"
#include "IterativeRefinementMPI.hpp"
#include "BiCGStabMPI.hpp"
#include "mpi.h"

namespace strumpack {

  /*! \brief This is the fully distributed solver.
   *
   * All steps are distributed: the symbolic factorization, the
   * factorization and the solve. The sparse factors (fill-in) are
   * distributed over the MPI processes. Only the MC64 phase is not
   * distributed. If MC64 reordering is enabled, the sparse input
   * matrix will be gathered to a single process, where then MC64 is
   * called. This can be a bottleneck for large matrices. Try to
   * disable MC64, since for many matrices, this reordering is not
   * required.
   *
   * The StrumpackSparseSolverMPIDist class depends on 3 template
   * paramaters <scalar_t,real_t,integer_t>. Supported combinations
   * for scalar_t,real_t are: <float,float>, <double,double>,
   * <std::complex<float>,float> and
   * <std::complex<double>,double>. The integer_t type defaults to a
   * regular int. If regular int causes 32 bit integer overflows, you
   * should switch to integer_t=int64_t instead.
   */
  template<typename scalar_t,typename real_t,typename integer_t>
  class StrumpackSparseSolverMPIDist : public StrumpackSparseSolverMPI<scalar_t,real_t,integer_t> {
  public:
    /*! \brief Constructor taking an MPI communicator and command line
     *         arguments.
     *
     * \param mpi_comm MPI communicator. Can be MPI_COMM_WORLD or a subcommunicator.
     * \param argc  The number of arguments, i.e, number of elements in the argv array.
     * \param argv  Command line arguments. Add -h or --help to have a description printed.
     * \param verb  Flag to suppres/enable output. Only the root of mpi_comm will print output to stdout.
     */
    StrumpackSparseSolverMPIDist(MPI_Comm mpi_comm, int argc, char* argv[], bool verb=true);
    virtual ~StrumpackSparseSolverMPIDist() {}
    void set_matrix(CSRMatrix<scalar_t,real_t,integer_t>& A);
    void set_matrix(CSCMatrix<scalar_t,real_t,integer_t>& A);
    /*! \brief Associate a (distributed) CSRMatrix with this solver.
     *
     * This matrix will not be modified. An internal copy will be
     * made, so it is safe to delete the data immediately after
     * calling this function. When this is called on a
     * StrumpackSparseSolverMPIDist object, the input matrix will
     * immediately be distributed over all the processes in the
     * communicator associated with the solver. This method is thus
     * collective on the MPI communicator associated with the solver.
     *
     * \param A  A CSRMatrixMPI<scalar_t,real_t,integer_t> object.
     */
    void set_matrix(CSRMatrixMPI<scalar_t,real_t,integer_t>& A);
    void set_csr_matrix(integer_t N, integer_t* row_ptr, integer_t* col_ind,
			scalar_t* values, bool symmetric_pattern=false);
    void set_csc_matrix(integer_t N, integer_t* col_ptr, integer_t* row_ind,
			scalar_t* values, bool symmetric_pattern=false);
    /*! \brief Associate a block-row distributed CSR matrix with the solver object.
     *
     * \param local_rows The number of rows of the input matrix assigned to this MPI process.
     *                   This should equal dist[rank+1]-dist[rank].
     * \param row_ptr    Indices in col_ind and values for the start of each row.
     *                   Nonzeros for row r+dist[rank] are in [row_ptr[r],row_ptr[r+1]).
     * \param col_ind    Column indices of each nonzero.
     * \param values     Nonzero values. Should have at least (row_ptr[dist[p+1]-dist[p]]-row_ptr[0]) elements.
     * \param dist       Specifies the block-row distribution.
     *                   A process with rank p owns rows [dist[p],dist[p+1]).
     * \param symmetric_pattern Denotes whether the sparsity pattern of the input matrix is symmetric.
     */
    void set_distributed_csr_matrix(integer_t local_rows, integer_t* row_ptr, integer_t* col_ind,
				    scalar_t* values, integer_t* dist, bool symmetric_pattern=false);
    /*! \brief Associate a (PETSc) MPIAIJ block-row distributed CSR matrix with the solver object. */
    void set_MPIAIJ_matrix(integer_t local_rows, integer_t* d_ptr, integer_t* d_ind, scalar_t* d_val,
			   integer_t* o_ptr, integer_t* o_ind, scalar_t* o_val, integer_t* garray);
    params::returnCode reorder(int nx=1, int ny=1, int nz=1);
    params::returnCode solve(scalar_t* b, scalar_t* x, bool use_initial_guess=false);

  protected:
    EliminationTree<scalar_t,real_t,integer_t>* get_elimination_tree();
    MatrixReordering<scalar_t,real_t,integer_t>* get_matrix_reordering();
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::StrumpackSparseSolverMPIDist
  (MPI_Comm mpi_comm, int argc, char* argv[], bool verb) :
    StrumpackSparseSolverMPI<scalar_t,real_t,integer_t>(mpi_comm, argc, argv, verb) {
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_matrix
  (CSRMatrix<scalar_t,real_t,integer_t>& A) {
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>(&A, this->comm, true);
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_matrix
  (CSCMatrix<scalar_t,real_t,integer_t>& A) {
    auto mat_csr_seq = new CSRMatrix<scalar_t,real_t,integer_t>(A);
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>(mat_csr_seq, this->comm, true);
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_matrix
  (CSRMatrixMPI<scalar_t,real_t,integer_t>& A) {
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>(A);
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_csr_matrix
  (integer_t N, integer_t* row_ptr, integer_t* col_ind,
   scalar_t* values, bool symmetric_pattern) {
    auto mat_seq = new CSRMatrix<scalar_t,real_t,integer_t>(N, row_ptr, col_ind, values, symmetric_pattern);
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>(mat_seq, this->comm, true);
    delete mat_seq;
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_csc_matrix
  (integer_t N, integer_t* col_ptr, integer_t* row_ind, scalar_t* values, bool symmetric_pattern) {
    auto mat_seq_csc = new CSCMatrix<scalar_t,real_t,integer_t>(N, col_ptr, row_ind, values, symmetric_pattern);
    auto mat_seq_csr = new CSRMatrix<scalar_t,real_t,integer_t>(*mat_seq_csc);
    delete mat_seq_csc;
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>(mat_seq_csr, this->comm, true);
    delete mat_seq_csr;
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_distributed_csr_matrix
  (integer_t local_rows, integer_t* row_ptr, integer_t* col_ind,
   scalar_t* values, integer_t* dist, bool symmetric_pattern) {
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>
      (local_rows, row_ptr, col_ind, values, dist, this->comm, symmetric_pattern);
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::set_MPIAIJ_matrix
  (integer_t local_rows, integer_t* d_ptr, integer_t* d_ind, scalar_t* d_val,
   integer_t* o_ptr, integer_t* o_ind, scalar_t* o_val, integer_t* garray) {
    delete this->mat;
    this->mat = new CSRMatrixMPI<scalar_t,real_t,integer_t>
      (local_rows, d_ptr, d_ind, d_val, o_ptr, o_ind, o_val, garray, this->comm);
    this->factored = this->reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  EliminationTree<scalar_t,real_t,integer_t>*
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::get_elimination_tree() {
    auto mat_dist = dynamic_cast<CSRMatrixMPI<scalar_t,real_t,integer_t>*>(this->mat);
    auto nd_dist = dynamic_cast<MatrixReorderingMPI<scalar_t,real_t,integer_t>*>(this->nd);
    assert(mat_dist != NULL);
    assert(nd_dist != NULL);
    return new EliminationTreeMPIDist<scalar_t,real_t,integer_t>
      (mat_dist, nd_dist, this->hss_strategy.use_hss, this->comm);
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  MatrixReordering<scalar_t,real_t,integer_t>*
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::
  get_matrix_reordering() {
    MatrixReordering<scalar_t,real_t,integer_t>* mr = NULL;
    switch (this->matrix_reordering) {
    case params::METIS: { mr = new ParMetisReordering<scalar_t,real_t,integer_t>(this->comm); break; }
    case params::SCOTCH: { mr = new PTScotchReordering<scalar_t,real_t,integer_t>(this->comm); break; }
    case params::GEOMETRIC: { mr = new GeometricReorderingMPI<scalar_t,real_t,integer_t>(this->comm); break; }
    default: { mr = StrumpackSparseSolverMPI<scalar_t,real_t,integer_t>::get_matrix_reordering(); break; }
    }
    return mr;
  }

  template<typename scalar_t,typename real_t,typename integer_t> params::returnCode
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::reorder(int nx, int ny, int nz) {
    MPI_Pcontrol(1, "reorder");
    auto ret = StrumpackSparseSolver<scalar_t,real_t,integer_t>::reorder(nx, ny, nz);
    MPI_Pcontrol(-1, "reorder");
    return ret;
  }

  /* TODO b is modified here! perhaps avoid that? */
  template<typename scalar_t,typename real_t,typename integer_t> params::returnCode
  StrumpackSparseSolverMPIDist<scalar_t,real_t,integer_t>::solve
  (scalar_t* b, scalar_t* x, bool use_initial_guess) {
    auto mat_mpi = dynamic_cast<CSRMatrixMPI<scalar_t,real_t,integer_t>*>(this->mat);
    assert(mat_mpi != NULL);

    if (!this->factored) {
      params::returnCode ierr = this->factor();
      if (ierr) return ierr;
    }
    TaskTimer t("solve");
    this->perf_counters_start();
    t.start();
    auto n_local = mat_mpi->local_rows();
    this->et->allocate_solve_work_memory();
    this->Krylov_iterations = 0;

    if (params::mc64job == 5) x_mult_y(n_local, b, this->mc64_Dr.data());

    auto gmres = [&](std::function<void(scalar_t*)> prec) {
      GMResMPI<scalar_t,real_t,integer_t>
      (this->comm, mat_mpi, prec, n_local, x, b,
       this->relative_Krylov_tolerance, this->absolute_Krylov_tolerance,
       this->Krylov_iterations, this->maximum_Krylov_iterations,
       this->gmres_restart, this->Gram_Schmidt_type, use_initial_guess, this->verbose && this->is_root);
    };
    auto bicgstab = [&](std::function<void(scalar_t*)> prec) {
      BiCGStabMPI<scalar_t,real_t,integer_t>
      (this->comm, mat_mpi, prec, n_local, x, b, this->relative_Krylov_tolerance, this->absolute_Krylov_tolerance,
       this->Krylov_iterations, this->maximum_Krylov_iterations, use_initial_guess, this->verbose && this->is_root);
    };
    auto prec = [&](scalar_t* x) {
      //MPI_Pcontrol(1, "multifrontal_solve_dist");
      this->et->multifrontal_solve_dist(x, mat_mpi->get_dist());
      //MPI_Pcontrol(-1, "multifrontal_solve_dist");
    };
    auto refine = [&]() {
      IterativeRefinementMPI<scalar_t,real_t,integer_t>
      (this->comm, mat_mpi, prec, n_local, x, b, this->relative_Krylov_tolerance, this->absolute_Krylov_tolerance,
       this->Krylov_iterations, this->maximum_Krylov_iterations, use_initial_guess, this->verbose && this->is_root);
    };

    switch (this->Krylov_solver) {
    case params::AUTO: {
      if (this->hss_strategy.use_hss) gmres(prec);
      else refine();
    }; break;
    case params::REFINE:     { refine(); }; break;
    case params::GMRES:      { gmres([](scalar_t*){}); }; break;
    case params::PREC_GMRES: { gmres(prec);  }; break;
    case params::BICGSTAB:      { bicgstab([](scalar_t*){}); }; break;
    case params::PREC_BICGSTAB: { bicgstab(prec);  }; break;
    case params::DIRECT: {
      std::copy(b, b+n_local, x);
      prec(x);
    }; break;
    }

    if (params::mc64job != 0)
      permute_vector(x, this->mc64_cperm, mat_mpi->get_dist(), this->comm);
    if (params::mc64job == 5) {
      x_mult_y(n_local, x, this->mc64_Dc.data());
      x_div_y(n_local, b, this->mc64_Dr.data());
    }
    // if (this->verbose) {
    //   auto max_res = mat_mpi->max_scaled_residual(x, b);
    //   if (this->is_root) std::cout << "MAX SCALED RES = " << max_res << std::endl;
    // }

    this->et->delete_solve_work_memory();
    t.stop();
    this->perf_counters_stop("DIRECT/GMRES solve");
    this->print_solve_stats(t);
    return params::SUCCESS;
  }

} // end namespace strumpack

#endif
